"""Extensions module, for backwards compatibility."""

# pylint: disable=unused-import
from pandera.api.extensions import (
    register_builtin_check,
    register_builtin_hypothesis,
    CheckType,
    register_check_method,
    register_check_statistics,
)
